# This file is part of Viper - https://github.com/botherder/viper
# See the file 'LICENSE' for copying permission.

import os
import hashlib
import binascii

try:
    import pydeep
    HAVE_SSDEEP = True
except ImportError:
    HAVE_SSDEEP = False

try:
    import magic
except ImportError:
    pass

class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class File(object):

    def __init__(self, path):
        self.path = path
        self.name = ''
        self.size = 0
        self.type = ''
        self.mime = ''
        self.md5 = ''
        self.sha1 = ''
        self.sha256 = ''
        self.sha512 = ''
        self.crc32 = ''
        self.ssdeep = ''
        self.tags = ''
        
        if self.is_valid():
            self.name = os.path.basename(self.path)
            self.size = os.path.getsize(self.path)
            self.type = self.get_type()
            self.mime = self.get_mime()
            self.get_hashes()
            self.ssdeep = self.get_ssdeep()

    @property
    def data(self):
        return open(self.path, 'rb').read()

    def is_valid(self):
        return os.path.exists(self.path) and os.path.isfile(self.path)# and os.path.getsize(self.path) != 0

    def get_chunks(self):
        fd = open(self.path, 'rb')
        while True:
            chunk = fd.read(16 * 1024)
            if not chunk:
                break

            yield chunk

        fd.close()

    def get_hashes(self):
        crc = 0
        md5 = hashlib.md5()
        sha1 = hashlib.sha1()
        sha256 = hashlib.sha256()
        sha512 = hashlib.sha512()
        
        for chunk in self.get_chunks():
            crc = binascii.crc32(chunk, crc)
            md5.update(chunk)
            sha1.update(chunk)
            sha256.update(chunk)
            sha512.update(chunk)

        self.crc32 = ''.join('%02X' % ((crc>>i)&0xff) for i in [24, 16, 8, 0])
        self.md5 = md5.hexdigest()
        self.sha1 = sha1.hexdigest()
        self.sha256 = sha256.hexdigest()
        self.sha512 = sha512.hexdigest()

    def get_ssdeep(self):
        if not HAVE_SSDEEP:
            return ''

        try:
            return pydeep.hash_file(self.path)
        except Exception:
            return ''

    def get_type(self):
        try:
            ms = magic.open(magic.MAGIC_NONE)
            ms.load()
            file_type = ms.file(self.path)
        except:
            try:
                file_type = magic.from_file(self.path)
            except:
                try:
                    import subprocess
                    file_process = subprocess.Popen(['file', '-b', self.path], stdout = subprocess.PIPE)
                    file_type = file_process.stdout.read().strip()
                except:
                    return ''
        finally:
            try:
                ms.close()
            except:
                pass

        return file_type

    def get_mime(self):
        try:
            ms = magic.open(magic.MIME)
            ms.load()
            mime_type = ms.file(self.path)
        except:
            try:
                mime = magic.Magic(mime=True)
                mime_type = mime.from_file(self.path)
            except:
                return ''

        return mime_type
